#!/usr/bin/perl -w
######################################################################
#
# $Id: FTimes-EadRoutines.t,v 1.8 2013/02/14 16:55:19 mavrik Exp $
#
######################################################################
#
# Copyright 2008-2013 The FTimes Project, All Rights Reserved.
#
######################################################################
#
# Purpose: Tests for FTimes::EadRoutines.
#
######################################################################

use 5.008;
use strict;
use Test;

BEGIN
{
  my %hPlan =
  (
    'tests' => (1 + (8 * 256)),
  );
  plan(%hPlan);
};

######################################################################
#
# Tests
#
######################################################################

  ####################################################################
  #
  # Test: The target module must load without error.
  #
  ####################################################################

  use FTimes::EadRoutines;
  ok(1);

  ####################################################################
  #
  # Common variables.
  #
  ####################################################################

  my (%hTargetValues);

  ####################################################################
  #
  # Test: URL encode 4-byte strings ranging from '%00' to '%ff'.
  #
  ####################################################################

  %hTargetValues =
  (
    '0x00' => '%00%00%00%00',
    '0x01' => '%01%01%01%01',
    '0x02' => '%02%02%02%02',
    '0x03' => '%03%03%03%03',
    '0x04' => '%04%04%04%04',
    '0x05' => '%05%05%05%05',
    '0x06' => '%06%06%06%06',
    '0x07' => '%07%07%07%07',
    '0x08' => '%08%08%08%08',
    '0x09' => '%09%09%09%09',
    '0x0a' => '%0a%0a%0a%0a',
    '0x0b' => '%0b%0b%0b%0b',
    '0x0c' => '%0c%0c%0c%0c',
    '0x0d' => '%0d%0d%0d%0d',
    '0x0e' => '%0e%0e%0e%0e',
    '0x0f' => '%0f%0f%0f%0f',
    '0x10' => '%10%10%10%10',
    '0x11' => '%11%11%11%11',
    '0x12' => '%12%12%12%12',
    '0x13' => '%13%13%13%13',
    '0x14' => '%14%14%14%14',
    '0x15' => '%15%15%15%15',
    '0x16' => '%16%16%16%16',
    '0x17' => '%17%17%17%17',
    '0x18' => '%18%18%18%18',
    '0x19' => '%19%19%19%19',
    '0x1a' => '%1a%1a%1a%1a',
    '0x1b' => '%1b%1b%1b%1b',
    '0x1c' => '%1c%1c%1c%1c',
    '0x1d' => '%1d%1d%1d%1d',
    '0x1e' => '%1e%1e%1e%1e',
    '0x1f' => '%1f%1f%1f%1f',
    '0x20' => '++++',
    '0x21' => '!!!!',
    '0x22' => '%22%22%22%22',
    '0x23' => '%23%23%23%23',
    '0x24' => '$$$$',
    '0x25' => '%25%25%25%25',
    '0x26' => '&&&&',
    '0x27' => '%27%27%27%27',
    '0x28' => '((((',
    '0x29' => '))))',
    '0x2a' => '****',
    '0x2b' => '%2b%2b%2b%2b',
    '0x2c' => ',,,,',
    '0x2d' => '----',
    '0x2e' => '....',
    '0x2f' => '////',
    '0x30' => '0000',
    '0x31' => '1111',
    '0x32' => '2222',
    '0x33' => '3333',
    '0x34' => '4444',
    '0x35' => '5555',
    '0x36' => '6666',
    '0x37' => '7777',
    '0x38' => '8888',
    '0x39' => '9999',
    '0x3a' => '::::',
    '0x3b' => ';;;;',
    '0x3c' => '<<<<',
    '0x3d' => '====',
    '0x3e' => '>>>>',
    '0x3f' => '????',
    '0x40' => '@@@@',
    '0x41' => 'AAAA',
    '0x42' => 'BBBB',
    '0x43' => 'CCCC',
    '0x44' => 'DDDD',
    '0x45' => 'EEEE',
    '0x46' => 'FFFF',
    '0x47' => 'GGGG',
    '0x48' => 'HHHH',
    '0x49' => 'IIII',
    '0x4a' => 'JJJJ',
    '0x4b' => 'KKKK',
    '0x4c' => 'LLLL',
    '0x4d' => 'MMMM',
    '0x4e' => 'NNNN',
    '0x4f' => 'OOOO',
    '0x50' => 'PPPP',
    '0x51' => 'QQQQ',
    '0x52' => 'RRRR',
    '0x53' => 'SSSS',
    '0x54' => 'TTTT',
    '0x55' => 'UUUU',
    '0x56' => 'VVVV',
    '0x57' => 'WWWW',
    '0x58' => 'XXXX',
    '0x59' => 'YYYY',
    '0x5a' => 'ZZZZ',
    '0x5b' => '[[[[',
    '0x5c' => '\\' x 4, # Need to escape the escapes.
    '0x5d' => ']]]]',
    '0x5e' => '^^^^',
    '0x5f' => '____',
    '0x60' => '%60%60%60%60',
    '0x61' => 'aaaa',
    '0x62' => 'bbbb',
    '0x63' => 'cccc',
    '0x64' => 'dddd',
    '0x65' => 'eeee',
    '0x66' => 'ffff',
    '0x67' => 'gggg',
    '0x68' => 'hhhh',
    '0x69' => 'iiii',
    '0x6a' => 'jjjj',
    '0x6b' => 'kkkk',
    '0x6c' => 'llll',
    '0x6d' => 'mmmm',
    '0x6e' => 'nnnn',
    '0x6f' => 'oooo',
    '0x70' => 'pppp',
    '0x71' => 'qqqq',
    '0x72' => 'rrrr',
    '0x73' => 'ssss',
    '0x74' => 'tttt',
    '0x75' => 'uuuu',
    '0x76' => 'vvvv',
    '0x77' => 'wwww',
    '0x78' => 'xxxx',
    '0x79' => 'yyyy',
    '0x7a' => 'zzzz',
    '0x7b' => '{{{{',
    '0x7c' => '%7c%7c%7c%7c',
    '0x7d' => '}}}}',
    '0x7e' => '~~~~',
    '0x7f' => '%7f%7f%7f%7f',
    '0x80' => '%80%80%80%80',
    '0x81' => '%81%81%81%81',
    '0x82' => '%82%82%82%82',
    '0x83' => '%83%83%83%83',
    '0x84' => '%84%84%84%84',
    '0x85' => '%85%85%85%85',
    '0x86' => '%86%86%86%86',
    '0x87' => '%87%87%87%87',
    '0x88' => '%88%88%88%88',
    '0x89' => '%89%89%89%89',
    '0x8a' => '%8a%8a%8a%8a',
    '0x8b' => '%8b%8b%8b%8b',
    '0x8c' => '%8c%8c%8c%8c',
    '0x8d' => '%8d%8d%8d%8d',
    '0x8e' => '%8e%8e%8e%8e',
    '0x8f' => '%8f%8f%8f%8f',
    '0x90' => '%90%90%90%90',
    '0x91' => '%91%91%91%91',
    '0x92' => '%92%92%92%92',
    '0x93' => '%93%93%93%93',
    '0x94' => '%94%94%94%94',
    '0x95' => '%95%95%95%95',
    '0x96' => '%96%96%96%96',
    '0x97' => '%97%97%97%97',
    '0x98' => '%98%98%98%98',
    '0x99' => '%99%99%99%99',
    '0x9a' => '%9a%9a%9a%9a',
    '0x9b' => '%9b%9b%9b%9b',
    '0x9c' => '%9c%9c%9c%9c',
    '0x9d' => '%9d%9d%9d%9d',
    '0x9e' => '%9e%9e%9e%9e',
    '0x9f' => '%9f%9f%9f%9f',
    '0xa0' => '%a0%a0%a0%a0',
    '0xa1' => '%a1%a1%a1%a1',
    '0xa2' => '%a2%a2%a2%a2',
    '0xa3' => '%a3%a3%a3%a3',
    '0xa4' => '%a4%a4%a4%a4',
    '0xa5' => '%a5%a5%a5%a5',
    '0xa6' => '%a6%a6%a6%a6',
    '0xa7' => '%a7%a7%a7%a7',
    '0xa8' => '%a8%a8%a8%a8',
    '0xa9' => '%a9%a9%a9%a9',
    '0xaa' => '%aa%aa%aa%aa',
    '0xab' => '%ab%ab%ab%ab',
    '0xac' => '%ac%ac%ac%ac',
    '0xad' => '%ad%ad%ad%ad',
    '0xae' => '%ae%ae%ae%ae',
    '0xaf' => '%af%af%af%af',
    '0xb0' => '%b0%b0%b0%b0',
    '0xb1' => '%b1%b1%b1%b1',
    '0xb2' => '%b2%b2%b2%b2',
    '0xb3' => '%b3%b3%b3%b3',
    '0xb4' => '%b4%b4%b4%b4',
    '0xb5' => '%b5%b5%b5%b5',
    '0xb6' => '%b6%b6%b6%b6',
    '0xb7' => '%b7%b7%b7%b7',
    '0xb8' => '%b8%b8%b8%b8',
    '0xb9' => '%b9%b9%b9%b9',
    '0xba' => '%ba%ba%ba%ba',
    '0xbb' => '%bb%bb%bb%bb',
    '0xbc' => '%bc%bc%bc%bc',
    '0xbd' => '%bd%bd%bd%bd',
    '0xbe' => '%be%be%be%be',
    '0xbf' => '%bf%bf%bf%bf',
    '0xc0' => '%c0%c0%c0%c0',
    '0xc1' => '%c1%c1%c1%c1',
    '0xc2' => '%c2%c2%c2%c2',
    '0xc3' => '%c3%c3%c3%c3',
    '0xc4' => '%c4%c4%c4%c4',
    '0xc5' => '%c5%c5%c5%c5',
    '0xc6' => '%c6%c6%c6%c6',
    '0xc7' => '%c7%c7%c7%c7',
    '0xc8' => '%c8%c8%c8%c8',
    '0xc9' => '%c9%c9%c9%c9',
    '0xca' => '%ca%ca%ca%ca',
    '0xcb' => '%cb%cb%cb%cb',
    '0xcc' => '%cc%cc%cc%cc',
    '0xcd' => '%cd%cd%cd%cd',
    '0xce' => '%ce%ce%ce%ce',
    '0xcf' => '%cf%cf%cf%cf',
    '0xd0' => '%d0%d0%d0%d0',
    '0xd1' => '%d1%d1%d1%d1',
    '0xd2' => '%d2%d2%d2%d2',
    '0xd3' => '%d3%d3%d3%d3',
    '0xd4' => '%d4%d4%d4%d4',
    '0xd5' => '%d5%d5%d5%d5',
    '0xd6' => '%d6%d6%d6%d6',
    '0xd7' => '%d7%d7%d7%d7',
    '0xd8' => '%d8%d8%d8%d8',
    '0xd9' => '%d9%d9%d9%d9',
    '0xda' => '%da%da%da%da',
    '0xdb' => '%db%db%db%db',
    '0xdc' => '%dc%dc%dc%dc',
    '0xdd' => '%dd%dd%dd%dd',
    '0xde' => '%de%de%de%de',
    '0xdf' => '%df%df%df%df',
    '0xe0' => '%e0%e0%e0%e0',
    '0xe1' => '%e1%e1%e1%e1',
    '0xe2' => '%e2%e2%e2%e2',
    '0xe3' => '%e3%e3%e3%e3',
    '0xe4' => '%e4%e4%e4%e4',
    '0xe5' => '%e5%e5%e5%e5',
    '0xe6' => '%e6%e6%e6%e6',
    '0xe7' => '%e7%e7%e7%e7',
    '0xe8' => '%e8%e8%e8%e8',
    '0xe9' => '%e9%e9%e9%e9',
    '0xea' => '%ea%ea%ea%ea',
    '0xeb' => '%eb%eb%eb%eb',
    '0xec' => '%ec%ec%ec%ec',
    '0xed' => '%ed%ed%ed%ed',
    '0xee' => '%ee%ee%ee%ee',
    '0xef' => '%ef%ef%ef%ef',
    '0xf0' => '%f0%f0%f0%f0',
    '0xf1' => '%f1%f1%f1%f1',
    '0xf2' => '%f2%f2%f2%f2',
    '0xf3' => '%f3%f3%f3%f3',
    '0xf4' => '%f4%f4%f4%f4',
    '0xf5' => '%f5%f5%f5%f5',
    '0xf6' => '%f6%f6%f6%f6',
    '0xf7' => '%f7%f7%f7%f7',
    '0xf8' => '%f8%f8%f8%f8',
    '0xf9' => '%f9%f9%f9%f9',
    '0xfa' => '%fa%fa%fa%fa',
    '0xfb' => '%fb%fb%fb%fb',
    '0xfc' => '%fc%fc%fc%fc',
    '0xfd' => '%fd%fd%fd%fd',
    '0xfe' => '%fe%fe%fe%fe',
    '0xff' => '%ff%ff%ff%ff',
  );

  foreach my $sNumber (0..255)
  {
    my $sHexNumber = sprintf("0x%02x", $sNumber);
    my $sActualValue = EadFTimesUrlEncode(chr($sNumber) x 4);
    my $sTargetValue = $hTargetValues{$sHexNumber};
    ok($sActualValue eq $sTargetValue);
  }

  ####################################################################
  #
  # Test: URL decode 4-byte strings ranging from '%00' to '%ff'.
  #
  ####################################################################

  foreach my $sNumber (0..255)
  {
    my $sHexNumber = sprintf("0x%02x", $sNumber);
    my $sActualValue = EadFTimesUrlDecode($hTargetValues{$sHexNumber});
    my $sTargetValue = chr($sNumber) x 4;
    ok($sActualValue eq $sTargetValue);
  }

  ####################################################################
  #
  # Test: NSS encode 4-byte strings ranging from '%00' to '%ff'.
  #
  ####################################################################

  %hTargetValues =
  (
    '0x00' => '_00__00__00__00_',
    '0x01' => '_01__01__01__01_',
    '0x02' => '_02__02__02__02_',
    '0x03' => '_03__03__03__03_',
    '0x04' => '_04__04__04__04_',
    '0x05' => '_05__05__05__05_',
    '0x06' => '_06__06__06__06_',
    '0x07' => '_07__07__07__07_',
    '0x08' => '_08__08__08__08_',
    '0x09' => '_09__09__09__09_',
    '0x0a' => '_0a__0a__0a__0a_',
    '0x0b' => '_0b__0b__0b__0b_',
    '0x0c' => '_0c__0c__0c__0c_',
    '0x0d' => '_0d__0d__0d__0d_',
    '0x0e' => '_0e__0e__0e__0e_',
    '0x0f' => '_0f__0f__0f__0f_',
    '0x10' => '_10__10__10__10_',
    '0x11' => '_11__11__11__11_',
    '0x12' => '_12__12__12__12_',
    '0x13' => '_13__13__13__13_',
    '0x14' => '_14__14__14__14_',
    '0x15' => '_15__15__15__15_',
    '0x16' => '_16__16__16__16_',
    '0x17' => '_17__17__17__17_',
    '0x18' => '_18__18__18__18_',
    '0x19' => '_19__19__19__19_',
    '0x1a' => '_1a__1a__1a__1a_',
    '0x1b' => '_1b__1b__1b__1b_',
    '0x1c' => '_1c__1c__1c__1c_',
    '0x1d' => '_1d__1d__1d__1d_',
    '0x1e' => '_1e__1e__1e__1e_',
    '0x1f' => '_1f__1f__1f__1f_',
    '0x20' => '_20__20__20__20_',
    '0x21' => '_21__21__21__21_',
    '0x22' => '_22__22__22__22_',
    '0x23' => '_23__23__23__23_',
    '0x24' => '_24__24__24__24_',
    '0x25' => '_25__25__25__25_',
    '0x26' => '_26__26__26__26_',
    '0x27' => '_27__27__27__27_',
    '0x28' => '_28__28__28__28_',
    '0x29' => '_29__29__29__29_',
    '0x2a' => '_2a__2a__2a__2a_',
    '0x2b' => '_2b__2b__2b__2b_',
    '0x2c' => '_2c__2c__2c__2c_',
    '0x2d' => '----',
    '0x2e' => '....',
    '0x2f' => '////',
    '0x30' => '0000',
    '0x31' => '1111',
    '0x32' => '2222',
    '0x33' => '3333',
    '0x34' => '4444',
    '0x35' => '5555',
    '0x36' => '6666',
    '0x37' => '7777',
    '0x38' => '8888',
    '0x39' => '9999',
    '0x3a' => '_3a__3a__3a__3a_',
    '0x3b' => '_3b__3b__3b__3b_',
    '0x3c' => '_3c__3c__3c__3c_',
    '0x3d' => '_3d__3d__3d__3d_',
    '0x3e' => '_3e__3e__3e__3e_',
    '0x3f' => '_3f__3f__3f__3f_',
    '0x40' => '_40__40__40__40_',
    '0x41' => 'AAAA',
    '0x42' => 'BBBB',
    '0x43' => 'CCCC',
    '0x44' => 'DDDD',
    '0x45' => 'EEEE',
    '0x46' => 'FFFF',
    '0x47' => 'GGGG',
    '0x48' => 'HHHH',
    '0x49' => 'IIII',
    '0x4a' => 'JJJJ',
    '0x4b' => 'KKKK',
    '0x4c' => 'LLLL',
    '0x4d' => 'MMMM',
    '0x4e' => 'NNNN',
    '0x4f' => 'OOOO',
    '0x50' => 'PPPP',
    '0x51' => 'QQQQ',
    '0x52' => 'RRRR',
    '0x53' => 'SSSS',
    '0x54' => 'TTTT',
    '0x55' => 'UUUU',
    '0x56' => 'VVVV',
    '0x57' => 'WWWW',
    '0x58' => 'XXXX',
    '0x59' => 'YYYY',
    '0x5a' => 'ZZZZ',
    '0x5b' => '_5b__5b__5b__5b_',
    '0x5c' => '_5c__5c__5c__5c_',
    '0x5d' => '_5d__5d__5d__5d_',
    '0x5e' => '_5e__5e__5e__5e_',
    '0x5f' => '_5f__5f__5f__5f_',
    '0x60' => '_60__60__60__60_',
    '0x61' => 'aaaa',
    '0x62' => 'bbbb',
    '0x63' => 'cccc',
    '0x64' => 'dddd',
    '0x65' => 'eeee',
    '0x66' => 'ffff',
    '0x67' => 'gggg',
    '0x68' => 'hhhh',
    '0x69' => 'iiii',
    '0x6a' => 'jjjj',
    '0x6b' => 'kkkk',
    '0x6c' => 'llll',
    '0x6d' => 'mmmm',
    '0x6e' => 'nnnn',
    '0x6f' => 'oooo',
    '0x70' => 'pppp',
    '0x71' => 'qqqq',
    '0x72' => 'rrrr',
    '0x73' => 'ssss',
    '0x74' => 'tttt',
    '0x75' => 'uuuu',
    '0x76' => 'vvvv',
    '0x77' => 'wwww',
    '0x78' => 'xxxx',
    '0x79' => 'yyyy',
    '0x7a' => 'zzzz',
    '0x7b' => '_7b__7b__7b__7b_',
    '0x7c' => '_7c__7c__7c__7c_',
    '0x7d' => '_7d__7d__7d__7d_',
    '0x7e' => '_7e__7e__7e__7e_',
    '0x7f' => '_7f__7f__7f__7f_',
    '0x80' => '_80__80__80__80_',
    '0x81' => '_81__81__81__81_',
    '0x82' => '_82__82__82__82_',
    '0x83' => '_83__83__83__83_',
    '0x84' => '_84__84__84__84_',
    '0x85' => '_85__85__85__85_',
    '0x86' => '_86__86__86__86_',
    '0x87' => '_87__87__87__87_',
    '0x88' => '_88__88__88__88_',
    '0x89' => '_89__89__89__89_',
    '0x8a' => '_8a__8a__8a__8a_',
    '0x8b' => '_8b__8b__8b__8b_',
    '0x8c' => '_8c__8c__8c__8c_',
    '0x8d' => '_8d__8d__8d__8d_',
    '0x8e' => '_8e__8e__8e__8e_',
    '0x8f' => '_8f__8f__8f__8f_',
    '0x90' => '_90__90__90__90_',
    '0x91' => '_91__91__91__91_',
    '0x92' => '_92__92__92__92_',
    '0x93' => '_93__93__93__93_',
    '0x94' => '_94__94__94__94_',
    '0x95' => '_95__95__95__95_',
    '0x96' => '_96__96__96__96_',
    '0x97' => '_97__97__97__97_',
    '0x98' => '_98__98__98__98_',
    '0x99' => '_99__99__99__99_',
    '0x9a' => '_9a__9a__9a__9a_',
    '0x9b' => '_9b__9b__9b__9b_',
    '0x9c' => '_9c__9c__9c__9c_',
    '0x9d' => '_9d__9d__9d__9d_',
    '0x9e' => '_9e__9e__9e__9e_',
    '0x9f' => '_9f__9f__9f__9f_',
    '0xa0' => '_a0__a0__a0__a0_',
    '0xa1' => '_a1__a1__a1__a1_',
    '0xa2' => '_a2__a2__a2__a2_',
    '0xa3' => '_a3__a3__a3__a3_',
    '0xa4' => '_a4__a4__a4__a4_',
    '0xa5' => '_a5__a5__a5__a5_',
    '0xa6' => '_a6__a6__a6__a6_',
    '0xa7' => '_a7__a7__a7__a7_',
    '0xa8' => '_a8__a8__a8__a8_',
    '0xa9' => '_a9__a9__a9__a9_',
    '0xaa' => '_aa__aa__aa__aa_',
    '0xab' => '_ab__ab__ab__ab_',
    '0xac' => '_ac__ac__ac__ac_',
    '0xad' => '_ad__ad__ad__ad_',
    '0xae' => '_ae__ae__ae__ae_',
    '0xaf' => '_af__af__af__af_',
    '0xb0' => '_b0__b0__b0__b0_',
    '0xb1' => '_b1__b1__b1__b1_',
    '0xb2' => '_b2__b2__b2__b2_',
    '0xb3' => '_b3__b3__b3__b3_',
    '0xb4' => '_b4__b4__b4__b4_',
    '0xb5' => '_b5__b5__b5__b5_',
    '0xb6' => '_b6__b6__b6__b6_',
    '0xb7' => '_b7__b7__b7__b7_',
    '0xb8' => '_b8__b8__b8__b8_',
    '0xb9' => '_b9__b9__b9__b9_',
    '0xba' => '_ba__ba__ba__ba_',
    '0xbb' => '_bb__bb__bb__bb_',
    '0xbc' => '_bc__bc__bc__bc_',
    '0xbd' => '_bd__bd__bd__bd_',
    '0xbe' => '_be__be__be__be_',
    '0xbf' => '_bf__bf__bf__bf_',
    '0xc0' => '_c0__c0__c0__c0_',
    '0xc1' => '_c1__c1__c1__c1_',
    '0xc2' => '_c2__c2__c2__c2_',
    '0xc3' => '_c3__c3__c3__c3_',
    '0xc4' => '_c4__c4__c4__c4_',
    '0xc5' => '_c5__c5__c5__c5_',
    '0xc6' => '_c6__c6__c6__c6_',
    '0xc7' => '_c7__c7__c7__c7_',
    '0xc8' => '_c8__c8__c8__c8_',
    '0xc9' => '_c9__c9__c9__c9_',
    '0xca' => '_ca__ca__ca__ca_',
    '0xcb' => '_cb__cb__cb__cb_',
    '0xcc' => '_cc__cc__cc__cc_',
    '0xcd' => '_cd__cd__cd__cd_',
    '0xce' => '_ce__ce__ce__ce_',
    '0xcf' => '_cf__cf__cf__cf_',
    '0xd0' => '_d0__d0__d0__d0_',
    '0xd1' => '_d1__d1__d1__d1_',
    '0xd2' => '_d2__d2__d2__d2_',
    '0xd3' => '_d3__d3__d3__d3_',
    '0xd4' => '_d4__d4__d4__d4_',
    '0xd5' => '_d5__d5__d5__d5_',
    '0xd6' => '_d6__d6__d6__d6_',
    '0xd7' => '_d7__d7__d7__d7_',
    '0xd8' => '_d8__d8__d8__d8_',
    '0xd9' => '_d9__d9__d9__d9_',
    '0xda' => '_da__da__da__da_',
    '0xdb' => '_db__db__db__db_',
    '0xdc' => '_dc__dc__dc__dc_',
    '0xdd' => '_dd__dd__dd__dd_',
    '0xde' => '_de__de__de__de_',
    '0xdf' => '_df__df__df__df_',
    '0xe0' => '_e0__e0__e0__e0_',
    '0xe1' => '_e1__e1__e1__e1_',
    '0xe2' => '_e2__e2__e2__e2_',
    '0xe3' => '_e3__e3__e3__e3_',
    '0xe4' => '_e4__e4__e4__e4_',
    '0xe5' => '_e5__e5__e5__e5_',
    '0xe6' => '_e6__e6__e6__e6_',
    '0xe7' => '_e7__e7__e7__e7_',
    '0xe8' => '_e8__e8__e8__e8_',
    '0xe9' => '_e9__e9__e9__e9_',
    '0xea' => '_ea__ea__ea__ea_',
    '0xeb' => '_eb__eb__eb__eb_',
    '0xec' => '_ec__ec__ec__ec_',
    '0xed' => '_ed__ed__ed__ed_',
    '0xee' => '_ee__ee__ee__ee_',
    '0xef' => '_ef__ef__ef__ef_',
    '0xf0' => '_f0__f0__f0__f0_',
    '0xf1' => '_f1__f1__f1__f1_',
    '0xf2' => '_f2__f2__f2__f2_',
    '0xf3' => '_f3__f3__f3__f3_',
    '0xf4' => '_f4__f4__f4__f4_',
    '0xf5' => '_f5__f5__f5__f5_',
    '0xf6' => '_f6__f6__f6__f6_',
    '0xf7' => '_f7__f7__f7__f7_',
    '0xf8' => '_f8__f8__f8__f8_',
    '0xf9' => '_f9__f9__f9__f9_',
    '0xfa' => '_fa__fa__fa__fa_',
    '0xfb' => '_fb__fb__fb__fb_',
    '0xfc' => '_fc__fc__fc__fc_',
    '0xfd' => '_fd__fd__fd__fd_',
    '0xfe' => '_fe__fe__fe__fe_',
    '0xff' => '_ff__ff__ff__ff_',
  );

  foreach my $sNumber (0..255)
  {
    my $sHexNumber = sprintf("0x%02x", $sNumber);
    my $sActualValue = EadFTimesNssEncode(chr($sNumber) x 4);
    my $sTargetValue = $hTargetValues{$sHexNumber};
    ok($sActualValue eq $sTargetValue);
  }

  ####################################################################
  #
  # Test: URL decode 4-byte strings ranging from '%00' to '%ff'.
  #
  ####################################################################

  foreach my $sNumber (0..255)
  {
    my $sHexNumber = sprintf("0x%02x", $sNumber);
    my $sActualValue = EadFTimesNssDecode($hTargetValues{$sHexNumber});
    my $sTargetValue = chr($sNumber) x 4;
    ok($sActualValue eq $sTargetValue);
  }

  ####################################################################
  #
  # Test: SSS encode 4-byte strings ranging from '%00' to '%ff'.
  #
  ####################################################################

  %hTargetValues =
  (
    '0x00' => '%00%00%00%00',
    '0x01' => '%01%01%01%01',
    '0x02' => '%02%02%02%02',
    '0x03' => '%03%03%03%03',
    '0x04' => '%04%04%04%04',
    '0x05' => '%05%05%05%05',
    '0x06' => '%06%06%06%06',
    '0x07' => '%07%07%07%07',
    '0x08' => '%08%08%08%08',
    '0x09' => '%09%09%09%09',
    '0x0a' => '%0a%0a%0a%0a',
    '0x0b' => '%0b%0b%0b%0b',
    '0x0c' => '%0c%0c%0c%0c',
    '0x0d' => '%0d%0d%0d%0d',
    '0x0e' => '%0e%0e%0e%0e',
    '0x0f' => '%0f%0f%0f%0f',
    '0x10' => '%10%10%10%10',
    '0x11' => '%11%11%11%11',
    '0x12' => '%12%12%12%12',
    '0x13' => '%13%13%13%13',
    '0x14' => '%14%14%14%14',
    '0x15' => '%15%15%15%15',
    '0x16' => '%16%16%16%16',
    '0x17' => '%17%17%17%17',
    '0x18' => '%18%18%18%18',
    '0x19' => '%19%19%19%19',
    '0x1a' => '%1a%1a%1a%1a',
    '0x1b' => '%1b%1b%1b%1b',
    '0x1c' => '%1c%1c%1c%1c',
    '0x1d' => '%1d%1d%1d%1d',
    '0x1e' => '%1e%1e%1e%1e',
    '0x1f' => '%1f%1f%1f%1f',
    '0x20' => '++++',
    '0x21' => '%21%21%21%21',
    '0x22' => '%22%22%22%22',
    '0x23' => '%23%23%23%23',
    '0x24' => '%24%24%24%24',
    '0x25' => '%25%25%25%25',
    '0x26' => '%26%26%26%26',
    '0x27' => '%27%27%27%27',
    '0x28' => '%28%28%28%28',
    '0x29' => '%29%29%29%29',
    '0x2a' => '%2a%2a%2a%2a',
    '0x2b' => '%2b%2b%2b%2b',
    '0x2c' => ',,,,',
    '0x2d' => '----',
    '0x2e' => '....',
    '0x2f' => '////',
    '0x30' => '0000',
    '0x31' => '1111',
    '0x32' => '2222',
    '0x33' => '3333',
    '0x34' => '4444',
    '0x35' => '5555',
    '0x36' => '6666',
    '0x37' => '7777',
    '0x38' => '8888',
    '0x39' => '9999',
    '0x3a' => '::::',
    '0x3b' => '%3b%3b%3b%3b',
    '0x3c' => '%3c%3c%3c%3c',
    '0x3d' => '%3d%3d%3d%3d',
    '0x3e' => '%3e%3e%3e%3e',
    '0x3f' => '%3f%3f%3f%3f',
    '0x40' => '@@@@',
    '0x41' => 'AAAA',
    '0x42' => 'BBBB',
    '0x43' => 'CCCC',
    '0x44' => 'DDDD',
    '0x45' => 'EEEE',
    '0x46' => 'FFFF',
    '0x47' => 'GGGG',
    '0x48' => 'HHHH',
    '0x49' => 'IIII',
    '0x4a' => 'JJJJ',
    '0x4b' => 'KKKK',
    '0x4c' => 'LLLL',
    '0x4d' => 'MMMM',
    '0x4e' => 'NNNN',
    '0x4f' => 'OOOO',
    '0x50' => 'PPPP',
    '0x51' => 'QQQQ',
    '0x52' => 'RRRR',
    '0x53' => 'SSSS',
    '0x54' => 'TTTT',
    '0x55' => 'UUUU',
    '0x56' => 'VVVV',
    '0x57' => 'WWWW',
    '0x58' => 'XXXX',
    '0x59' => 'YYYY',
    '0x5a' => 'ZZZZ',
    '0x5b' => '%5b%5b%5b%5b',
    '0x5c' => '%5c%5c%5c%5c',
    '0x5d' => '%5d%5d%5d%5d',
    '0x5e' => '%5e%5e%5e%5e',
    '0x5f' => '____',
    '0x60' => '%60%60%60%60',
    '0x61' => 'aaaa',
    '0x62' => 'bbbb',
    '0x63' => 'cccc',
    '0x64' => 'dddd',
    '0x65' => 'eeee',
    '0x66' => 'ffff',
    '0x67' => 'gggg',
    '0x68' => 'hhhh',
    '0x69' => 'iiii',
    '0x6a' => 'jjjj',
    '0x6b' => 'kkkk',
    '0x6c' => 'llll',
    '0x6d' => 'mmmm',
    '0x6e' => 'nnnn',
    '0x6f' => 'oooo',
    '0x70' => 'pppp',
    '0x71' => 'qqqq',
    '0x72' => 'rrrr',
    '0x73' => 'ssss',
    '0x74' => 'tttt',
    '0x75' => 'uuuu',
    '0x76' => 'vvvv',
    '0x77' => 'wwww',
    '0x78' => 'xxxx',
    '0x79' => 'yyyy',
    '0x7a' => 'zzzz',
    '0x7b' => '%7b%7b%7b%7b',
    '0x7c' => '%7c%7c%7c%7c',
    '0x7d' => '%7d%7d%7d%7d',
    '0x7e' => '%7e%7e%7e%7e',
    '0x7f' => '%7f%7f%7f%7f',
    '0x80' => '%80%80%80%80',
    '0x81' => '%81%81%81%81',
    '0x82' => '%82%82%82%82',
    '0x83' => '%83%83%83%83',
    '0x84' => '%84%84%84%84',
    '0x85' => '%85%85%85%85',
    '0x86' => '%86%86%86%86',
    '0x87' => '%87%87%87%87',
    '0x88' => '%88%88%88%88',
    '0x89' => '%89%89%89%89',
    '0x8a' => '%8a%8a%8a%8a',
    '0x8b' => '%8b%8b%8b%8b',
    '0x8c' => '%8c%8c%8c%8c',
    '0x8d' => '%8d%8d%8d%8d',
    '0x8e' => '%8e%8e%8e%8e',
    '0x8f' => '%8f%8f%8f%8f',
    '0x90' => '%90%90%90%90',
    '0x91' => '%91%91%91%91',
    '0x92' => '%92%92%92%92',
    '0x93' => '%93%93%93%93',
    '0x94' => '%94%94%94%94',
    '0x95' => '%95%95%95%95',
    '0x96' => '%96%96%96%96',
    '0x97' => '%97%97%97%97',
    '0x98' => '%98%98%98%98',
    '0x99' => '%99%99%99%99',
    '0x9a' => '%9a%9a%9a%9a',
    '0x9b' => '%9b%9b%9b%9b',
    '0x9c' => '%9c%9c%9c%9c',
    '0x9d' => '%9d%9d%9d%9d',
    '0x9e' => '%9e%9e%9e%9e',
    '0x9f' => '%9f%9f%9f%9f',
    '0xa0' => '%a0%a0%a0%a0',
    '0xa1' => '%a1%a1%a1%a1',
    '0xa2' => '%a2%a2%a2%a2',
    '0xa3' => '%a3%a3%a3%a3',
    '0xa4' => '%a4%a4%a4%a4',
    '0xa5' => '%a5%a5%a5%a5',
    '0xa6' => '%a6%a6%a6%a6',
    '0xa7' => '%a7%a7%a7%a7',
    '0xa8' => '%a8%a8%a8%a8',
    '0xa9' => '%a9%a9%a9%a9',
    '0xaa' => '%aa%aa%aa%aa',
    '0xab' => '%ab%ab%ab%ab',
    '0xac' => '%ac%ac%ac%ac',
    '0xad' => '%ad%ad%ad%ad',
    '0xae' => '%ae%ae%ae%ae',
    '0xaf' => '%af%af%af%af',
    '0xb0' => '%b0%b0%b0%b0',
    '0xb1' => '%b1%b1%b1%b1',
    '0xb2' => '%b2%b2%b2%b2',
    '0xb3' => '%b3%b3%b3%b3',
    '0xb4' => '%b4%b4%b4%b4',
    '0xb5' => '%b5%b5%b5%b5',
    '0xb6' => '%b6%b6%b6%b6',
    '0xb7' => '%b7%b7%b7%b7',
    '0xb8' => '%b8%b8%b8%b8',
    '0xb9' => '%b9%b9%b9%b9',
    '0xba' => '%ba%ba%ba%ba',
    '0xbb' => '%bb%bb%bb%bb',
    '0xbc' => '%bc%bc%bc%bc',
    '0xbd' => '%bd%bd%bd%bd',
    '0xbe' => '%be%be%be%be',
    '0xbf' => '%bf%bf%bf%bf',
    '0xc0' => '%c0%c0%c0%c0',
    '0xc1' => '%c1%c1%c1%c1',
    '0xc2' => '%c2%c2%c2%c2',
    '0xc3' => '%c3%c3%c3%c3',
    '0xc4' => '%c4%c4%c4%c4',
    '0xc5' => '%c5%c5%c5%c5',
    '0xc6' => '%c6%c6%c6%c6',
    '0xc7' => '%c7%c7%c7%c7',
    '0xc8' => '%c8%c8%c8%c8',
    '0xc9' => '%c9%c9%c9%c9',
    '0xca' => '%ca%ca%ca%ca',
    '0xcb' => '%cb%cb%cb%cb',
    '0xcc' => '%cc%cc%cc%cc',
    '0xcd' => '%cd%cd%cd%cd',
    '0xce' => '%ce%ce%ce%ce',
    '0xcf' => '%cf%cf%cf%cf',
    '0xd0' => '%d0%d0%d0%d0',
    '0xd1' => '%d1%d1%d1%d1',
    '0xd2' => '%d2%d2%d2%d2',
    '0xd3' => '%d3%d3%d3%d3',
    '0xd4' => '%d4%d4%d4%d4',
    '0xd5' => '%d5%d5%d5%d5',
    '0xd6' => '%d6%d6%d6%d6',
    '0xd7' => '%d7%d7%d7%d7',
    '0xd8' => '%d8%d8%d8%d8',
    '0xd9' => '%d9%d9%d9%d9',
    '0xda' => '%da%da%da%da',
    '0xdb' => '%db%db%db%db',
    '0xdc' => '%dc%dc%dc%dc',
    '0xdd' => '%dd%dd%dd%dd',
    '0xde' => '%de%de%de%de',
    '0xdf' => '%df%df%df%df',
    '0xe0' => '%e0%e0%e0%e0',
    '0xe1' => '%e1%e1%e1%e1',
    '0xe2' => '%e2%e2%e2%e2',
    '0xe3' => '%e3%e3%e3%e3',
    '0xe4' => '%e4%e4%e4%e4',
    '0xe5' => '%e5%e5%e5%e5',
    '0xe6' => '%e6%e6%e6%e6',
    '0xe7' => '%e7%e7%e7%e7',
    '0xe8' => '%e8%e8%e8%e8',
    '0xe9' => '%e9%e9%e9%e9',
    '0xea' => '%ea%ea%ea%ea',
    '0xeb' => '%eb%eb%eb%eb',
    '0xec' => '%ec%ec%ec%ec',
    '0xed' => '%ed%ed%ed%ed',
    '0xee' => '%ee%ee%ee%ee',
    '0xef' => '%ef%ef%ef%ef',
    '0xf0' => '%f0%f0%f0%f0',
    '0xf1' => '%f1%f1%f1%f1',
    '0xf2' => '%f2%f2%f2%f2',
    '0xf3' => '%f3%f3%f3%f3',
    '0xf4' => '%f4%f4%f4%f4',
    '0xf5' => '%f5%f5%f5%f5',
    '0xf6' => '%f6%f6%f6%f6',
    '0xf7' => '%f7%f7%f7%f7',
    '0xf8' => '%f8%f8%f8%f8',
    '0xf9' => '%f9%f9%f9%f9',
    '0xfa' => '%fa%fa%fa%fa',
    '0xfb' => '%fb%fb%fb%fb',
    '0xfc' => '%fc%fc%fc%fc',
    '0xfd' => '%fd%fd%fd%fd',
    '0xfe' => '%fe%fe%fe%fe',
    '0xff' => '%ff%ff%ff%ff',
  );

  foreach my $sNumber (0..255)
  {
    my $sHexNumber = sprintf("0x%02x", $sNumber);
    my $sActualValue = EadFTimesSssEncode(chr($sNumber) x 4);
    my $sTargetValue = $hTargetValues{$sHexNumber};
    ok($sActualValue eq $sTargetValue);
  }

  ####################################################################
  #
  # Test: SSS decode 4-byte strings ranging from '%00' to '%ff'.
  #
  ####################################################################

  foreach my $sNumber (0..255)
  {
    my $sHexNumber = sprintf("0x%02x", $sNumber);
    my $sActualValue = EadFTimesSssDecode($hTargetValues{$sHexNumber});
    my $sTargetValue = chr($sNumber) x 4;
    ok($sActualValue eq $sTargetValue);
  }

  ####################################################################
  #
  # Test: Base64 encode 4-byte strings ranging from '%00' to '%ff'.
  #
  ####################################################################

  %hTargetValues =
  (
    '0x00' => 'AAAAAA==',
    '0x01' => 'AQEBAQ==',
    '0x02' => 'AgICAg==',
    '0x03' => 'AwMDAw==',
    '0x04' => 'BAQEBA==',
    '0x05' => 'BQUFBQ==',
    '0x06' => 'BgYGBg==',
    '0x07' => 'BwcHBw==',
    '0x08' => 'CAgICA==',
    '0x09' => 'CQkJCQ==',
    '0x0a' => 'CgoKCg==',
    '0x0b' => 'CwsLCw==',
    '0x0c' => 'DAwMDA==',
    '0x0d' => 'DQ0NDQ==',
    '0x0e' => 'Dg4ODg==',
    '0x0f' => 'Dw8PDw==',
    '0x10' => 'EBAQEA==',
    '0x11' => 'EREREQ==',
    '0x12' => 'EhISEg==',
    '0x13' => 'ExMTEw==',
    '0x14' => 'FBQUFA==',
    '0x15' => 'FRUVFQ==',
    '0x16' => 'FhYWFg==',
    '0x17' => 'FxcXFw==',
    '0x18' => 'GBgYGA==',
    '0x19' => 'GRkZGQ==',
    '0x1a' => 'GhoaGg==',
    '0x1b' => 'GxsbGw==',
    '0x1c' => 'HBwcHA==',
    '0x1d' => 'HR0dHQ==',
    '0x1e' => 'Hh4eHg==',
    '0x1f' => 'Hx8fHw==',
    '0x20' => 'ICAgIA==',
    '0x21' => 'ISEhIQ==',
    '0x22' => 'IiIiIg==',
    '0x23' => 'IyMjIw==',
    '0x24' => 'JCQkJA==',
    '0x25' => 'JSUlJQ==',
    '0x26' => 'JiYmJg==',
    '0x27' => 'JycnJw==',
    '0x28' => 'KCgoKA==',
    '0x29' => 'KSkpKQ==',
    '0x2a' => 'KioqKg==',
    '0x2b' => 'KysrKw==',
    '0x2c' => 'LCwsLA==',
    '0x2d' => 'LS0tLQ==',
    '0x2e' => 'Li4uLg==',
    '0x2f' => 'Ly8vLw==',
    '0x30' => 'MDAwMA==',
    '0x31' => 'MTExMQ==',
    '0x32' => 'MjIyMg==',
    '0x33' => 'MzMzMw==',
    '0x34' => 'NDQ0NA==',
    '0x35' => 'NTU1NQ==',
    '0x36' => 'NjY2Ng==',
    '0x37' => 'Nzc3Nw==',
    '0x38' => 'ODg4OA==',
    '0x39' => 'OTk5OQ==',
    '0x3a' => 'Ojo6Og==',
    '0x3b' => 'Ozs7Ow==',
    '0x3c' => 'PDw8PA==',
    '0x3d' => 'PT09PQ==',
    '0x3e' => 'Pj4+Pg==',
    '0x3f' => 'Pz8/Pw==',
    '0x40' => 'QEBAQA==',
    '0x41' => 'QUFBQQ==',
    '0x42' => 'QkJCQg==',
    '0x43' => 'Q0NDQw==',
    '0x44' => 'RERERA==',
    '0x45' => 'RUVFRQ==',
    '0x46' => 'RkZGRg==',
    '0x47' => 'R0dHRw==',
    '0x48' => 'SEhISA==',
    '0x49' => 'SUlJSQ==',
    '0x4a' => 'SkpKSg==',
    '0x4b' => 'S0tLSw==',
    '0x4c' => 'TExMTA==',
    '0x4d' => 'TU1NTQ==',
    '0x4e' => 'Tk5OTg==',
    '0x4f' => 'T09PTw==',
    '0x50' => 'UFBQUA==',
    '0x51' => 'UVFRUQ==',
    '0x52' => 'UlJSUg==',
    '0x53' => 'U1NTUw==',
    '0x54' => 'VFRUVA==',
    '0x55' => 'VVVVVQ==',
    '0x56' => 'VlZWVg==',
    '0x57' => 'V1dXVw==',
    '0x58' => 'WFhYWA==',
    '0x59' => 'WVlZWQ==',
    '0x5a' => 'WlpaWg==',
    '0x5b' => 'W1tbWw==',
    '0x5c' => 'XFxcXA==',
    '0x5d' => 'XV1dXQ==',
    '0x5e' => 'Xl5eXg==',
    '0x5f' => 'X19fXw==',
    '0x60' => 'YGBgYA==',
    '0x61' => 'YWFhYQ==',
    '0x62' => 'YmJiYg==',
    '0x63' => 'Y2NjYw==',
    '0x64' => 'ZGRkZA==',
    '0x65' => 'ZWVlZQ==',
    '0x66' => 'ZmZmZg==',
    '0x67' => 'Z2dnZw==',
    '0x68' => 'aGhoaA==',
    '0x69' => 'aWlpaQ==',
    '0x6a' => 'ampqag==',
    '0x6b' => 'a2traw==',
    '0x6c' => 'bGxsbA==',
    '0x6d' => 'bW1tbQ==',
    '0x6e' => 'bm5ubg==',
    '0x6f' => 'b29vbw==',
    '0x70' => 'cHBwcA==',
    '0x71' => 'cXFxcQ==',
    '0x72' => 'cnJycg==',
    '0x73' => 'c3Nzcw==',
    '0x74' => 'dHR0dA==',
    '0x75' => 'dXV1dQ==',
    '0x76' => 'dnZ2dg==',
    '0x77' => 'd3d3dw==',
    '0x78' => 'eHh4eA==',
    '0x79' => 'eXl5eQ==',
    '0x7a' => 'enp6eg==',
    '0x7b' => 'e3t7ew==',
    '0x7c' => 'fHx8fA==',
    '0x7d' => 'fX19fQ==',
    '0x7e' => 'fn5+fg==',
    '0x7f' => 'f39/fw==',
    '0x80' => 'gICAgA==',
    '0x81' => 'gYGBgQ==',
    '0x82' => 'goKCgg==',
    '0x83' => 'g4ODgw==',
    '0x84' => 'hISEhA==',
    '0x85' => 'hYWFhQ==',
    '0x86' => 'hoaGhg==',
    '0x87' => 'h4eHhw==',
    '0x88' => 'iIiIiA==',
    '0x89' => 'iYmJiQ==',
    '0x8a' => 'ioqKig==',
    '0x8b' => 'i4uLiw==',
    '0x8c' => 'jIyMjA==',
    '0x8d' => 'jY2NjQ==',
    '0x8e' => 'jo6Ojg==',
    '0x8f' => 'j4+Pjw==',
    '0x90' => 'kJCQkA==',
    '0x91' => 'kZGRkQ==',
    '0x92' => 'kpKSkg==',
    '0x93' => 'k5OTkw==',
    '0x94' => 'lJSUlA==',
    '0x95' => 'lZWVlQ==',
    '0x96' => 'lpaWlg==',
    '0x97' => 'l5eXlw==',
    '0x98' => 'mJiYmA==',
    '0x99' => 'mZmZmQ==',
    '0x9a' => 'mpqamg==',
    '0x9b' => 'm5ubmw==',
    '0x9c' => 'nJycnA==',
    '0x9d' => 'nZ2dnQ==',
    '0x9e' => 'np6eng==',
    '0x9f' => 'n5+fnw==',
    '0xa0' => 'oKCgoA==',
    '0xa1' => 'oaGhoQ==',
    '0xa2' => 'oqKiog==',
    '0xa3' => 'o6Ojow==',
    '0xa4' => 'pKSkpA==',
    '0xa5' => 'paWlpQ==',
    '0xa6' => 'pqampg==',
    '0xa7' => 'p6enpw==',
    '0xa8' => 'qKioqA==',
    '0xa9' => 'qampqQ==',
    '0xaa' => 'qqqqqg==',
    '0xab' => 'q6urqw==',
    '0xac' => 'rKysrA==',
    '0xad' => 'ra2trQ==',
    '0xae' => 'rq6urg==',
    '0xaf' => 'r6+vrw==',
    '0xb0' => 'sLCwsA==',
    '0xb1' => 'sbGxsQ==',
    '0xb2' => 'srKysg==',
    '0xb3' => 's7Ozsw==',
    '0xb4' => 'tLS0tA==',
    '0xb5' => 'tbW1tQ==',
    '0xb6' => 'tra2tg==',
    '0xb7' => 't7e3tw==',
    '0xb8' => 'uLi4uA==',
    '0xb9' => 'ubm5uQ==',
    '0xba' => 'urq6ug==',
    '0xbb' => 'u7u7uw==',
    '0xbc' => 'vLy8vA==',
    '0xbd' => 'vb29vQ==',
    '0xbe' => 'vr6+vg==',
    '0xbf' => 'v7+/vw==',
    '0xc0' => 'wMDAwA==',
    '0xc1' => 'wcHBwQ==',
    '0xc2' => 'wsLCwg==',
    '0xc3' => 'w8PDww==',
    '0xc4' => 'xMTExA==',
    '0xc5' => 'xcXFxQ==',
    '0xc6' => 'xsbGxg==',
    '0xc7' => 'x8fHxw==',
    '0xc8' => 'yMjIyA==',
    '0xc9' => 'ycnJyQ==',
    '0xca' => 'ysrKyg==',
    '0xcb' => 'y8vLyw==',
    '0xcc' => 'zMzMzA==',
    '0xcd' => 'zc3NzQ==',
    '0xce' => 'zs7Ozg==',
    '0xcf' => 'z8/Pzw==',
    '0xd0' => '0NDQ0A==',
    '0xd1' => '0dHR0Q==',
    '0xd2' => '0tLS0g==',
    '0xd3' => '09PT0w==',
    '0xd4' => '1NTU1A==',
    '0xd5' => '1dXV1Q==',
    '0xd6' => '1tbW1g==',
    '0xd7' => '19fX1w==',
    '0xd8' => '2NjY2A==',
    '0xd9' => '2dnZ2Q==',
    '0xda' => '2tra2g==',
    '0xdb' => '29vb2w==',
    '0xdc' => '3Nzc3A==',
    '0xdd' => '3d3d3Q==',
    '0xde' => '3t7e3g==',
    '0xdf' => '39/f3w==',
    '0xe0' => '4ODg4A==',
    '0xe1' => '4eHh4Q==',
    '0xe2' => '4uLi4g==',
    '0xe3' => '4+Pj4w==',
    '0xe4' => '5OTk5A==',
    '0xe5' => '5eXl5Q==',
    '0xe6' => '5ubm5g==',
    '0xe7' => '5+fn5w==',
    '0xe8' => '6Ojo6A==',
    '0xe9' => '6enp6Q==',
    '0xea' => '6urq6g==',
    '0xeb' => '6+vr6w==',
    '0xec' => '7Ozs7A==',
    '0xed' => '7e3t7Q==',
    '0xee' => '7u7u7g==',
    '0xef' => '7+/v7w==',
    '0xf0' => '8PDw8A==',
    '0xf1' => '8fHx8Q==',
    '0xf2' => '8vLy8g==',
    '0xf3' => '8/Pz8w==',
    '0xf4' => '9PT09A==',
    '0xf5' => '9fX19Q==',
    '0xf6' => '9vb29g==',
    '0xf7' => '9/f39w==',
    '0xf8' => '+Pj4+A==',
    '0xf9' => '+fn5+Q==',
    '0xfa' => '+vr6+g==',
    '0xfb' => '+/v7+w==',
    '0xfc' => '/Pz8/A==',
    '0xfd' => '/f39/Q==',
    '0xfe' => '/v7+/g==',
    '0xff' => '/////w==',
  );

  foreach my $sNumber (0..255)
  {
    my $sHexNumber = sprintf("0x%02x", $sNumber);
    my $sActualValue = EadBase64Encode(chr($sNumber) x 4);
    my $sTargetValue = $hTargetValues{$sHexNumber};
    ok($sActualValue eq $sTargetValue);
  }

  ####################################################################
  #
  # Test: Base64 decode 4-byte strings ranging from '%00' to '%ff'.
  #
  ####################################################################

  foreach my $sNumber (0..255)
  {
    my $sHexNumber = sprintf("0x%02x", $sNumber);
    my $sActualValue = EadBase64Decode($hTargetValues{$sHexNumber});
    my $sTargetValue = chr($sNumber) x 4;
    ok($sActualValue eq $sTargetValue);
  }

